--
--    Dynamic Moving PowerTakeOff
--
--    @author:      Wopster
--    @descripion: 
--    @website:     
--    @history:     v1.0 - 2016-12-13 - Initial implementation
--
--[[
    movingPtoFilenameStartNode = node where to load the pto from in the shared file. (Values: index of node)
    movingPtoInputNode = node where the pto should be linked todo. (Values: index of node)
    movingPtoAboveAttacher = is it above the attacherJoint? (Values: true/false)
    movingPtoVibrationOffset = how much should the pto vibrate? (Values: integer) / 1000
    movingPtoParkPosition = park position when detached. (Values: x y z trans)
    movingPtoParkRotation = park rotation of the attach node when detached. (Values: x y z rot)
    movingPtoRotSpeed = rotation speed. (Values: float)
    
    <inputAttacherJoint movingPtoInputNode="0>0|2" movingPtoAboveAttacher="false" movingPtoVibrationOffset="2" movingPtoParkPosition="0.1 -0.4 0.5" movingPtoParkRotation="50 0 0" />
]]

DynamicMovingPto = {
    debug = true,
    modDir = g_currentModDirectory
}

function DynamicMovingPto.prerequisitesPresent(specializations)
    return SpecializationUtil.hasSpecialization(AttacherJoints, specializations) and SpecializationUtil.hasSpecialization(PowerConsumer, specializations)
end

function DynamicMovingPto:load(savegame)
    self.loadInputMovingPowerTakeOff = DynamicMovingPto.loadInputMovingPowerTakeOff
    self.vibrateMovingPowerTakeOff = DynamicMovingPto.vibrateMovingPowerTakeOff
    self.attachMovingPowerTakeOff = DynamicMovingPto.attachMovingPowerTakeOff
    self.detachMovingPowerTakeOff = DynamicMovingPto.detachMovingPowerTakeOff
    self.updateAttacherJointGraphics = Utils.overwrittenFunction(self.updateAttacherJointGraphics, DynamicMovingPto.updatePowerTakeoff)
    
    local i = 0
    
    while true do
        local key = ('vehicle.inputAttacherJoints.inputAttacherJoint(%d)'):format(i)
        
        if not hasXMLProperty(self.xmlFile, key) then
            break
        end
        
        self:loadInputMovingPowerTakeOff(self.xmlFile, key, 'movingPto', self.inputAttacherJoints[i + 1])
        
        i = i + 1
    end
end

function DynamicMovingPto:loadInputMovingPowerTakeOff(xmlFile, key, type, entry)
    local ptoKey = ('%s#%s'):format(key, type)
    local inputNode = Utils.indexToObject(self.components, getXMLString(xmlFile, ptoKey .. 'InputNode'))
    
    if inputNode ~= nil then
        local movingPto = {
            isNonScalable = true,
            isParked = false,
            node = inputNode,
            aboveAttacher = Utils.getNoNil(getXMLBool(xmlFile, ptoKey .. 'AboveAttacher'), true),
            vibrationOffset = Utils.getNoNil(getXMLFloat(xmlFile, ptoKey .. 'VibrationOffset'), 1) / 1000,
            size = 0
        }
        
        local rotSpeed = math.rad(Utils.getNoNil(getXMLFloat(xmlFile, ptoKey .. 'RotSpeed'), 0) * 0.001)
        
        movingPto.rotSpeed = math.abs(rotSpeed) < 0.00001 and 0 or rotSpeed
        
        -- todo: load from local node?
        
        -- load from external filename
        local filename = Utils.getNoNil(getXMLString(xmlFile, ptoKey .. 'Filename'), 'shared/parts/pto.i3d')
        
        if filename ~= nil and filename ~= '' then
            local i3dNode = Utils.loadSharedI3DFile(filename, self.baseDirectory, false, false, false)
            
            if i3dNode ~= 0 then
                local dummyRoot = getChildAt(i3dNode, 0)
            
                if dummyRoot ~= 0 then
                    local startNode = getXMLFloat(xmlFile, ptoKey .. 'FilenameStartNode')
                    local parkPosition = Utils.getNoNil(Utils.getVectorNFromString(getXMLString(xmlFile, ptoKey .. 'ParkPosition'), 3), {0, 0, 0})
                    local parkRotation = Utils.getNoNil(Utils.getRadiansFromString(getXMLString(xmlFile, ptoKey .. 'ParkRotation'), 3), {0, 0, 0})
                    local ptoRpm = self.powerConsumer.ptoRpm ~= 0 and self.powerConsumer.ptoRpm or 540
                    local rootNode = startNode ~= nil and getChildAt(dummyRoot, startNode) or dummyRoot
                    
                    link(inputNode, rootNode)
                    delete(i3dNode)
                    setTranslation(rootNode, 0, 0, 0)
                    setRotation(rootNode, 0, math.rad(180), 0)
                    local rotationNode = getChildAt(rootNode, 0)
                    local linkNode = getChildAt(rootNode, 1)
                    local movingNode = getChildAt(rotationNode, 0)
                    local attachNode = getChildAt(movingNode, 0)
                    local attachRefNode = getChildAt(attachNode, 0)                                    
                    
                    local _, _, offsetRef = localToLocal(attachNode, attachRefNode, 0, 0, 0)                    
                    local ax, ay, az = getWorldTranslation(movingPto.node)
                    local bx, by, bz = getWorldTranslation(rotationNode)
                    movingPto.movingOffset = Utils.vector3Length(ax-bx, ay-by, az-bz)
                    
                    movingPto.rootNode = rootNode
                    movingPto.rotationNode = rotationNode
                    movingPto.movingNode = movingNode
                    movingPto.attachNode = attachNode
                    movingPto.linkNode = linkNode
                    movingPto.orgAttachTrans = {getTranslation(attachNode)}
                    movingPto.orgMovingTrans = {getTranslation(movingNode)}
                    movingPto.parkPosition = parkPosition
                    movingPto.parkRotation = parkRotation
                    movingPto.attachOffset = math.abs(offsetRef)
                    movingPto.rotationNodes = {
                        isRunning = false,
                        nodes = {
                            {
                                shaderComponentScale = {
                                    1, 0, 0, 0
                                },
                                currentRot = 0,
                                isActive = true,
                                node = linkNode,
                                rotSpeed = math.abs(math.rad(ptoRpm)),
                                rotAxis = 3,
                                turnOnFadeTime = 3000,
                                turnOffFadeTime = 3000,
                                turnOnOffCurValue = 0
                            },
                            {
                                shaderComponentScale = {
                                    1, 0, 0, 0
                                },
                                currentRot = 0,
                                isActive = true,
                                node = attachNode,
                                rotSpeed = math.abs(math.rad(ptoRpm)),
                                rotAxis = 3,
                                turnOnFadeTime = 3000,
                                turnOffFadeTime = 3000,
                                turnOnOffCurValue = 0
                            }
                        }                        
                    }
                    
                    DynamicMovingPto:parkMovingPowerTakeOff(movingPto)
                end
            end
        end
        
        entry[('%sInput'):format(type)] = movingPto
        entry[('%sActive'):format(type)] = false
    end
end

function DynamicMovingPto:delete()
    if self.movingPto ~= nil and self.movingPto.rootNode ~= nil then
        delete(self.movingPto.rootNode)
        delete(self.movingPto.rotationNode)       
        delete(self.movingPto.movingNode)       
        delete(self.movingPto.attachNode) 
        delete(self.movingPto.linkNode)    
    end
end

function DynamicMovingPto:mouseEvent(posX, posY, isDown, isUp, button)
end

function DynamicMovingPto:keyEvent(unicode, sym, modifier, isDown)
end

function DynamicMovingPto:postUpdate(dt)
    if self.attacherVehicle ~= nil then    
        if self.attacherVehicle.forcePtoUpdate then
            self.attacherVehicle:updatePowerTakeoff(self.attacherVehicle:getImplementByObject(self), dt, 'movingPto', 'pto')
        end
    end
end

function DynamicMovingPto:update(dt)
    if self.isClient then
        if self.movingPtoInput ~= nil then            
            Utils.updateRotationNodes(self, self.movingPtoInput.rotationNodes, dt, self:getIsActive() and self:getIsTurnedOn())

            if self.movingPtoInput.rotationNode ~= nil then
                DynamicMovingPto:vibrateMovingPowerTakeOff(self.movingPtoInput.rotationNode, self.movingPtoInput.vibrationOffset, dt, self:getIsActive() and self:getIsTurnedOn())
            end
        end
    end
end

function DynamicMovingPto:draw()
end

function DynamicMovingPto:onPreAttach(attacherVehicle, inputAttacherJointDescIndex)
    self.movingPtoInput = self.inputAttacherJoints[inputAttacherJointDescIndex].movingPtoInput
end

function DynamicMovingPto:vibrateMovingPowerTakeOff(node, offset, dt, isTurnedOn)
    if isTurnedOn then
        local trans = {getTranslation(node)}
        
        if trans[1] < offset and trans[2] < offset then
            trans[1] = trans[1] + (offset * 2)
            trans[2] = trans[2] + (offset * 2)
        else            
            trans[1] = trans[1] - (offset * 2)
            trans[2] = trans[2] - (offset * 2)
        end        
        
        setTranslation(node, unpack(trans))
    end
end

function DynamicMovingPto:attachMovingPowerTakeOff(object, input, outputJointDesc)
    link(outputJointDesc.node, input.attachNode)
    setTranslation(input.attachNode, 0, 0, input.attachOffset)
    setRotation(input.attachNode, math.rad(180), 0, 0)
    
    if object.addWashableNode ~= nil then
        object:addWashableNode(input.rotationNode) -- it's a merged node
        object:setDirtAmount(object:getDirtAmount(), true)
    end
    
    input.isParked = false
end

function DynamicMovingPto:detachMovingPowerTakeOff(object, input)
    unlink(input.attachNode)
    link(input.movingNode, input.attachNode)
                        
    if object.removeWashableNode ~= nil then
        object:removeWashableNode(input.rotationNode) -- it's a merged node
    end
    
    DynamicMovingPto:parkMovingPowerTakeOff(input)
end

function DynamicMovingPto:parkMovingPowerTakeOff(input)
    if not input.isParked then
        setTranslation(input.attachNode, unpack(input.orgAttachTrans))
        setRotation(input.attachNode, unpack(input.parkRotation))
        setTranslation(input.movingNode, unpack(input.orgMovingTrans))                        
        
        local direction = {localToLocal(input.attachNode, input.rootNode, unpack(input.parkPosition))}
        setDirection(input.rotationNode, direction[1], direction[2], direction[3], 0, 1, 0)
        setTranslation(input.movingNode, 0, 0, direction[3] + input.movingOffset)
                
        input.isParked = true
    end
end

function DynamicMovingPto:updatePowerTakeoff(superFunc, childFunc, implement, dt, type)    
    local object = implement.object
    
    if object.movingPtoInput == nil and type ~= 'movingPto' then
        if superFunc ~= nil then
            superFunc(self, childFunc, implement, dt, type)
        end
    end

    local jointDesc = self.attacherJoints[implement.jointDescIndex]
    local overwriteOutputType = 'pto'
    
    if overwriteOutputType ~= nil and jointDesc[('%sActive'):format(type)] then
        local output = jointDesc[('%sOutput'):format(overwriteOutputType)]
        local input = object[('%sInput'):format(type)]
    
        if input ~= nil then
            if input.isNonScalable ~= nil and input.isNonScalable then
                if input.rotSpeed ~= 0 and object:getIsPowerTakeoffActive() then
                    output.rotation = (output.rotation + dt * input.rotSpeed) % (2 * math.pi)
                    
                    setRotation(input.rotationNode, 0, 0, output.rotation)
                    setRotation(input.attachNode, 0, 0, output.rotation)
                end
                
                local rotationTrans = {getWorldTranslation(input.rotationNode)}
                local attachTrans = {getWorldTranslation(input.attachNode)}        
                local direction = {worldDirectionToLocal(getParent(input.rotationNode), attachTrans[1] - rotationTrans[1], attachTrans[2] - rotationTrans[2], attachTrans[3]- rotationTrans[3])}
                local distance = Utils.vector3Length(unpack(direction))
                
                setDirection(input.rotationNode, direction[1], direction[2], direction[3], 0, 1, 0)                
                setTranslation(input.movingNode, 0, 0, distance)
            end
        end
    end
end

function DynamicMovingPto:updateAttacherJointGraphics(superFunc, implement, dt)
    self:updatePowerTakeoff(implement, dt, 'movingPto', 'pto')
end

function DynamicMovingPto:onAttached(attacherVehicle, implement)
    if not g_currentMission.dynamicHoseIsManual then
        local jointDesc = attacherVehicle.attacherJoints[implement.jointDescIndex]
         
        if implement.object ~= nil then
            local attachPto = function(object, jointDesc, type, overwriteOutputType)
                if not jointDesc[('%sActive'):format(type)] then
                    local input = object[('%sInput'):format(type)]
                    
                    local outputJointDesc = jointDesc[('%sOutput'):format(overwriteOutputType)]
                    local outputObject = object[('%sOutput'):format(overwriteOutputType)]
                    
                    if input ~= nil then
                        local isActive = false
                        
                        if input.rootNode ~= nil then
                            if input.isNonScalable ~= nil and input.isNonScalable then
                                DynamicMovingPto:attachMovingPowerTakeOff(object, input, outputJointDesc)
                            end
                            
                            isActive = true
                        end
                
                        jointDesc[('%sActive'):format(type)] = isActive
                    end            
                end
            end
            
            attachPto(implement.object, jointDesc, 'movingPto', 'pto')
        end
    end
end

function DynamicMovingPto:onDetached(attacherVehicle, jointDescIndex)
    if not g_currentMission.dynamicHoseIsManual then
        local jointDesc = attacherVehicle.attacherJoints[jointDescIndex]
        
        local detachPto = function(object, jointDesc, type, overwriteOutputType)
            if jointDesc[('%sActive'):format(type)] then
                local input = object[('%sInput'):format(type)]            
                local outputJointDesc = jointDesc[('%sOutput'):format(overwriteOutputType)]
                local outputObject = object[('%sOutput'):format(overwriteOutputType)]
                local isActive = true
                
                if input ~= nil and input.rootNode ~= nil then
                    if input.isNonScalable ~= nil and input.isNonScalable then                    
                        DynamicMovingPto:detachMovingPowerTakeOff(object, input)
                    end
                    
                    isActive = false
                end
            end
            
            jointDesc[('%sActive'):format(type)] = isActive
        end
        
        detachPto(self, jointDesc, 'movingPto', 'pto')
    end
end

AttacherJoints.updatePowerTakeoff = Utils.overwrittenFunction(AttacherJoints.updatePowerTakeoff, DynamicMovingPto.updatePowerTakeoff)
-- AttacherJoints.updateAttacherJointGraphics = Utils.appendedFunction(AttacherJoints.updateAttacherJointGraphics, DynamicMovingPto.updateAttacherJointGraphics)